import re

from queuing_hub.conn.base import BasePub
from queuing_hub.conn.aws import AwsPub
from queuing_hub.conn.gcp import GcpPub


class Publisher:

    def __init__(
        self, aws_profile_name=None,
        gcp_credential_path=None,
        gcp_project=None
    ):
        self._aws_pub = AwsPub(profile_name=aws_profile_name)
        self._gcp_pub = GcpPub(
            credential_path=gcp_credential_path,
            project=gcp_project
        )

        self.__connectors: list(BasePub) = [self._aws_pub, self._gcp_pub]
        self._topic_list = []
        for connector in self.__connectors:
            self._topic_list.extend(connector.topic_list)

    @property
    def topic_list(self) -> list:
        return self._topic_list

    def push(self, topic_list: list, body: str) -> list:
        message_ids = []
        connector: BasePub

        for topic in topic_list:
            connector = self._get_connector(topic)
            message_ids.append(connector.push(topic, body))
        return message_ids

    def _get_connector(self, topic: str) -> BasePub:
        if re.search(
            r'https://.+-.+-.+\.queue\.amazonaws\.com/[0-9]+/.+',
            topic
        ):
            connector = self._aws_pub
        elif re.search(
            r'projects/[a-z0-9-]+/topics/.+',
            topic
        ):
            connector = self._gcp_pub
        else:
            raise ValueError(f'Invalid topic: {topic}')

        return connector



import json
import re

from queuing_hub.conn.base import BaseSub
from queuing_hub.conn.aws import AwsSub
from queuing_hub.conn.gcp import GcpSub


class Subscriber:

    def __init__(
        self, aws_profile_name=None,
        gcp_credential_path=None, gcp_project=None
    ):
        self._aws_sub = AwsSub(profile_name=aws_profile_name)
        self._gcp_sub = GcpSub(
            credential_path=gcp_credential_path,
            project=gcp_project
        )

        self.__connectors: list(BaseSub) = [self._aws_sub, self._gcp_sub]
        self._sub_list = []
        for connector in self.__connectors:
            self._sub_list.extend(connector.sub_list)

    @property
    def sub_list(self) -> list:
        return self._sub_list

    def qsize(self) -> str:
        response = {}
        for connector in self.__connectors:
            response.update(connector.qsize())
        return json.dumps(response, indent=2)

    def is_empty(self, sub_list: list) -> str:
        response = {}
        for sub in sub_list:
            connector = self.__get_connector(sub)
            response[sub] = connector.is_empty(sub)
        return json.dumps(response, indent=2)

    def purge(self, sub_list: list) -> None:
        for sub in sub_list:
            connector = self._get_connector(sub)
            connector.purge(sub)

    def pull(self, sub_list: list, max_num: int, ack: bool = True) -> list:
        response = []
        connector: BaseSub

        for sub in sub_list:
            connector = self._get_connector(sub)
            response = connector.pull(sub=sub, max_num=max_num, ack=ack)
            if response != []:
                break

        return response

    def pull_nack(self, sub_list: list, max_num: int) -> list:
        return self.pull(sub_list=sub_list, max_num=max_num, ack=False)

    def _get_connector(self, sub_path: str) -> BaseSub:
        if re.search(
            r'https://.+-.+-.+\.queue\.amazonaws\.com/[0-9]+/.+',
            sub_path
        ):
            connector = self._aws_sub
        elif re.search(
            r'projects/[a-z0-9-]+/subscriptions/.+',
            sub_path
        ):
            connector = self._gcp_sub
        else:
            raise ValueError(f'Invalid subscription: {sub_path}')

        return connector



class BasePub():
    def __init__(self):
        pass

    def push(self, topic: str, body: str) -> None:
        pass


class BaseSub():
    def __init__(self):
        pass

    def qsize(self, sub_list: list = None) -> dict:
        pass

    def is_empty(self, sub: str) -> bool:
        pass

    def purge(self, sub: str) -> None:
        pass

    def pull(self, sub: str, max_num: int = 1, ack: bool = False) -> list:
        pass

    def _ack(self, sub: str, messages: list) -> None:
        pass



from queuing_hub.publisher import Publisher
from queuing_hub.subscriber import Subscriber


class Forwarder:

    def __init__(
        self, sub: str, topic: str, max_num: int = 1,
        aws_profile_name=None,
        gcp_credential_path=None,
        gcp_project=None
    ):
        self.publisher = Publisher(
            aws_profile_name=aws_profile_name,
            gcp_credential_path=gcp_credential_path,
            gcp_project=gcp_project
        )
        self.subscriber = Subscriber(
            aws_profile_name=aws_profile_name,
            gcp_credential_path=gcp_credential_path,
            gcp_project=gcp_project
        )
        self.topic = topic
        self.sub = sub
        self.max_num = max_num

    def transport(self, ack: bool = True) -> list:
        messages = self.subscriber.pull(
            sub_list=[self.sub],
            max_num=self.max_num,
            ack=ack
        )

        responses = []
        if not messages:
            return responses

        for message in messages:
            response = self.publisher.push(
                topic_list=[self.topic],
                body=message
            )
            responses.append(response)
        return responses

    def pass_through(self):
        return self.transport(ack=False)



import boto3

from queuing_hub.conn.base import BasePub, BaseSub


class AwsBase():

    def __init__(self, profile_name=None):
        session = boto3.Session(profile_name=profile_name)
        self._client = session.client('sqs')
        self._queue_list = self._client.list_queues()['QueueUrls']


class AwsPub(AwsBase, BasePub):

    def __init__(self, profile_name=None):
        AwsBase.__init__(self, profile_name=profile_name)
        BasePub.__init__(self)

    @property
    def topic_list(self) -> list:
        return self._queue_list

    def push(self, topic: str, body: str) -> dict:
        response = self._client.send_message(
            QueueUrl=topic,
            MessageBody=body
        )
        return response['MessageId']


class AwsSub(AwsBase, BaseSub):

    ATTRIBUTE_NAMES = [
        'ApproximateNumberOfMessages',
        # 'ApproximateNumberOfMessagesDelayed',
        # 'ApproximateNumberOfMessagesNotVisible',
        # 'DelaySeconds',
        # 'MessageRetentionPeriod',
        # 'ReceiveMessageWaitTimeSeconds',
        # 'VisibilityTimeout'
    ]

    def __init__(self, profile_name=None):
        AwsBase.__init__(self, profile_name=profile_name)
        BaseSub.__init__(self)

    @property
    def sub_list(self) -> list:
        return self._queue_list

    def qsize(self, sub_list: list = None) -> dict:
        response = {'aws': {}}
        if not sub_list:
            sub_list = self._queue_list

        for sub in sub_list:
            response['aws'][sub] = self._get_message_count(sub)

        return response

    def is_empty(self, sub: str) -> bool:
        return self._get_message_count(sub) == 0

    def purge(self, sub: str) -> None:
        self._client.purge_queue(QueueUrl=sub)

    def pull(self, sub: str, max_num: int = 1, ack: bool = False) -> list:
        response = self._client.receive_message(
            QueueUrl=sub,
            MaxNumberOfMessages=max_num
        )

        messages = response.get('Messages')

        if ack and messages:
            self._ack(sub, messages)

        return [message.get('Body') for message in messages]

    def _ack(self, sub: str, messages: list) -> None:
        receipt_handle_list = \
            [message['ReceiptHandle'] for message in messages]
        for receipt_handle in receipt_handle_list:
            self._client.delete_message(
                QueueUrl=sub,
                ReceiptHandle=receipt_handle
            )

    def _get_message_count(self, sub: str) -> int:
        attributes = self._get_attributes(sub, self.ATTRIBUTE_NAMES)
        return int(attributes[self.ATTRIBUTE_NAMES[0]])

    def _get_attributes(self, sub: str, attribute_names: str) -> dict:
        response = self._client.get_queue_attributes(
            QueueUrl=sub,
            AttributeNames=attribute_names
        )
        return response['Attributes']



import os
from datetime import datetime
from concurrent.futures import TimeoutError as FuturesTimeoutError

from google.cloud import pubsub_v1
from google.cloud.monitoring_v3 import query, MetricServiceClient
from google.oauth2.service_account import Credentials

from queuing_hub.conn.base import BasePub, BaseSub

PROJECT = os.environ['GCP_PROJECT']


class GcpPub(BasePub):

    def __init__(self, credential_path=None, project=None):
        super().__init__()

        if credential_path:
            credentials = Credentials.from_service_account_file(
                filename=credential_path
            )
        else:
            credentials = None

        if project:
            self._project = project
        else:
            self._project = PROJECT

        self._publisher = pubsub_v1.PublisherClient(credentials=credentials)
        self._pub_organizer = pubsub_v1.publisher.client.publisher_client \
            .PublisherClient(credentials=credentials)

        # topics
        project_path = self._pub_organizer.common_project_path(self._project)
        topics_obj = self._pub_organizer.list_topics(project=project_path)
        self._topic_list = [queue.name for queue in topics_obj]

    @property
    def topic_list(self) -> list:
        return self._topic_list

    def push(self, topic: str, body: str) -> None:
        future = self._publisher.publish(topic, body.encode())
        # future.add_done_callback(self._callback)
        return future.result()

    @staticmethod
    def _callback(future):
        message_id = future.result()
        print(f'MessageId {message_id} has published! ')


class GcpSub(BaseSub):

    TIMEOUT = 5.0
    METRIC_TYPE = 'pubsub.googleapis.com/subscription/num_undelivered_messages'

    def __init__(self, credential_path=None, project=None):
        super().__init__()

        if credential_path:
            credentials = Credentials.from_service_account_file(
                filename=credential_path
            )
        else:
            credentials = None

        if project:
            self._project = project
        else:
            self._project = PROJECT

        self._subscriber = pubsub_v1.SubscriberClient(credentials=credentials)
        self._sub_organizer = pubsub_v1.subscriber.client.subscriber_client \
            .SubscriberClient(credentials=credentials)

        # subscriptions
        project_path = self._sub_organizer.common_project_path(self._project)
        sub_obj = self._sub_organizer.list_subscriptions(project=project_path)
        self._sub_list = [queue.name for queue in sub_obj]

    def __del__(self):
        self._subscriber.close()

    @property
    def sub_list(self) -> list:
        return self._sub_list

    def qsize(self, sub_list: list = None) -> dict:
        response = {'gcp': {}}
        if not sub_list:
            sub_list = self._sub_list

        query_results = query.Query(
            client=MetricServiceClient(),
            project=self._project,
            metric_type=self.METRIC_TYPE,
            end_time=datetime.now(),
            minutes=2
            # if set 1 minute, we get nothing
            # while creating the latest metrics.
        )

        for result in self.__read_metric(query_results=query_results):
            response['gcp'][result['subscription']] = result['value']

        return response

    def is_empty(self, sub: str) -> bool:
        return self.qsize([sub])[sub] == 0

    def pull(self, sub: str, max_num: int = 1, ack: bool = False) -> list:
        messages = []
        response = self._sub_organizer.pull(
            request={
                'subscription': sub,
                "max_messages": max_num,
            }
        )

        for msg in response.received_messages:
            messages.append(msg.message.data.decode())

        if ack:
            self._ack(sub=sub, messages=response.received_messages)

        return messages

    def pull_streaming(self, sub: str) -> None:
        streaming_pull_future = self._subscriber.subscribe(
            sub,
            callback=self.__streaming_pull_callback
        )
        print(f"Listening for messages on {sub}..\n")

        with self._subscriber:
            try:
                streaming_pull_future.result(timeout=self.TIMEOUT)
            except FuturesTimeoutError:
                streaming_pull_future.cancel()

    def purge(self, sub: str) -> None:
        seek_request = pubsub_v1.types.pubsub_gapic_types \
            .SeekRequest(subscription=sub, time=datetime.now())
        self._sub_organizer.seek(request=seek_request)

    def _ack(self, sub: str, messages: list) -> None:
        ack_ids = [msg.ack_id for msg in messages]
        self._sub_organizer.acknowledge(
            request={
                "subscription": sub,
                "ack_ids": ack_ids,
            }
        )

    def __read_metric(self, query_results: query.Query) -> dict:
        for content in query_results:
            sub_id = content.resource.labels['subscription_id']
            sub = self._sub_organizer.subscription_path(self._project, sub_id)
            yield {
                'subscription': sub,
                'value': content.points[0].value.int64_value
            }

    def __streaming_pull_callback(self, message):
        print(f'Received {message.data.decode()}.')
        if message.attributes:
            print('Attributes:')
            for key in message.attributes:
                value = message.attributes.get(key)
                print(f'{key}: {value}')
        message.ack()
